'''
Test redirection behavior to invalid addresses
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum
import re
import os
import socket
import sys

Test.Summary = '''
Test redirection behavior to invalid addresses
'''

Test.ContinueOnFail = False

Test.Setup.Copy(os.path.join(Test.Variables.AtsTestToolsDir, 'tcp_client.py'))

dns = Test.MakeDNServer('dns')
# This record is used in each test case to get the initial redirect response from the origin that we will handle.
dnsRecords = {'iwillredirect.test': ['127.0.0.1']}

host = socket.gethostname()
ipv4addrs = set()
try:
    ipv4addrs = set([ip for (family, _, _, _, (ip, *_)) in socket.getaddrinfo(host, port=None) if socket.AF_INET == family])
except socket.gaierror:
    pass

ipv6addrs = set()
try:
    ipv6addrs = set(
        [
            "[{0}]".format(ip.split('%')[0])
            for (family, _, _, _, (ip, *_)) in socket.getaddrinfo(host, port=None)
            if socket.AF_INET6 == family and 'fe80' != ip[0:4]
        ])  # Skip link-local addresses.
except socket.gaierror:
    pass

origin = Test.MakeOriginServer('origin', ip='0.0.0.0')
ArbitraryTimestamp = '12345678'

# This is for cases when the content is actually fetched from the invalid address.
request_header = {
    'headers': ('GET / HTTP/1.1\r\n'
                'Host: *\r\n\r\n'),
    'timestamp': ArbitraryTimestamp,
    'body': ''
}
response_header = {
    'headers': ('HTTP/1.1 204 No Content\r\n'
                'Connection: close\r\n\r\n'),
    'timestamp': ArbitraryTimestamp,
    'body': ''
}
origin.addResponse('sessionfile.log', request_header, response_header)

# Map scenarios to trafficserver processes.
trafficservers = {}

data_dirname = 'generated_test_data'
data_path = os.path.join(Test.TestDirectory, data_dirname)
os.makedirs(data_path, exist_ok=True)


def normalizeForAutest(value):
    '''
    autest uses "test run" names to build file and directory names, so we must transform them in case there are incompatible or
    annoying characters.
    This means we can also use them in URLs.
    '''
    if not value:
        return None
    return re.sub(r'[^a-z0-9-]', '_', value, flags=re.I)


def makeTestCase(redirectTarget, expectedAction, scenario):
    '''
    Helper method that creates a "meta-test" from which autest generates a test case.

    :param redirectTarget: The target address of a redirect from origin to be handled.
    :param scenario: Defines the ACL to configure and the addresses to test.
    '''

    config = ','.join(':'.join(t) for t in sorted((addr.name.lower(), action.name.lower()) for (addr, action) in scenario.items()))

    normRedirectTarget = normalizeForAutest(redirectTarget)
    normConfig = normalizeForAutest(config)
    tr = Test.AddTestRun('With_Config_{0}_Redirect_to_{1}'.format(normConfig, normRedirectTarget))

    if trafficservers:
        tr.StillRunningAfter = origin
        tr.StillRunningAfter = dns
    else:
        tr.Processes.Default.StartBefore(origin)
        tr.Processes.Default.StartBefore(dns)

    if config not in trafficservers:
        trafficservers[config] = Test.MakeATSProcess('ts_{0}'.format(normConfig), enable_cache=False)
        trafficservers[config].Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http|dns|redirect',
                'proxy.config.http.number_of_redirections': 1,
                'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(dns.Variables.Port),
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.url_remap.remap_required': 0,
                'proxy.config.http.redirect.actions': config,
                'proxy.config.http.connect_attempts_timeout': 5,
                'proxy.config.http.connect_attempts_max_retries': 0,
            })
        tr.Processes.Default.StartBefore(trafficservers[config])
    else:
        tr.StillRunningAfter = trafficservers[config]

    testDomain = 'testdomain{0}.test'.format(normRedirectTarget)
    # The micro DNS server can't tell us whether it has a record of the domain already, so we use a dictionary to avoid duplicates.
    # We remove any surrounding brackets that are common to IPv6 addresses.
    if redirectTarget:
        dnsRecords[testDomain] = [redirectTarget.strip('[]')]

    # A GET request parameterized on the config and on the target.
    request_header = {
        'headers': ('GET /redirect?config={0}&target={1} HTTP/1.1\r\n'
                    'Host: *\r\n\r\n').format(normConfig, normRedirectTarget),
        'timestamp': ArbitraryTimestamp,
        'body': ''
    }
    # Returns a redirect to the test domain for the given target & the port number for the TS of the given config.
    response_header = {
        'headers':
            ('HTTP/1.1 307 Temporary Redirect\r\n'
             'Location: http://{0}:{1}/\r\n'
             'Connection: close\r\n\r\n').format(testDomain, origin.Variables.Port),
        'timestamp': ArbitraryTimestamp,
        'body': ''
    }
    origin.addResponse('sessionfile.log', request_header, response_header)

    # Generate the request data file.
    command_path = os.path.join(data_path, tr.Name)
    with open(command_path, 'w') as f:
        f.write(
            ('GET /redirect?config={0}&target={1} HTTP/1.1\r\n'
             'Host: iwillredirect.test:{2}\r\n\r\n').format(normConfig, normRedirectTarget, origin.Variables.Port))
    # Set the command with the appropriate URL.
    port = trafficservers[config].Variables.port
    dir_path = os.path.join(data_dirname, tr.Name)
    tr.Processes.Default.Command = \
        (f"bash -o pipefail -c '{sys.executable} tcp_client.py 127.0.0.1 {port} "
         f"{dir_path} | head -n 1'")
    tr.Processes.Default.ReturnCode = 0
    # Generate and set the 'gold file' to check stdout
    goldFilePath = os.path.join(data_path, '{0}.gold'.format(tr.Name))
    with open(goldFilePath, 'w') as f:
        f.write(expectedAction.value['expectedStatusLine'])
    tr.Processes.Default.Streams.stdout = goldFilePath


class AddressE(Enum):
    '''
    Classes of addresses are mapped to example addresses.
    '''
    Private = ('10.0.0.1', '[fc00::1]')
    Loopback = (['127.1.2.3'])  # [::1] is omitted here because it is likely overwritten by Self, and there are no others in IPv6.
    Multicast = ('224.1.2.3', '[ff42::]')
    Linklocal = ('169.254.0.1', '[fe80::]')
    Routable = ('72.30.35.10', '[2001:4998:58:1836::10]')  # Do not Follow redirects to these in an automated test.
    Self = ipv4addrs | ipv6addrs  # Addresses of this host.
    Default = None  # All addresses apply, nothing in particular to test.


class ActionE(Enum):
    # Title case because 'return' is a Python keyword.
    Return = {'config': 'return', 'expectedStatusLine': 'HTTP/1.1 307 Temporary Redirect\r\n'}
    Reject = {'config': 'reject', 'expectedStatusLine': 'HTTP/1.1 403 Forbidden\r\n'}
    Follow = {'config': 'follow', 'expectedStatusLine': 'HTTP/1.1 204 No Content\r\n'}

    # Added to test failure modes.
    Break = {'expectedStatusLine': 'HTTP/1.1 500 Cannot find server.\r\n'}


scenarios = [
    {
        # Follow to loopback, but alternately reject/return others.
        AddressE.Private: ActionE.Reject,
        AddressE.Loopback: ActionE.Follow,
        AddressE.Multicast: ActionE.Reject,
        AddressE.Linklocal: ActionE.Return,
        AddressE.Routable: ActionE.Reject,
        AddressE.Self: ActionE.Return,
        AddressE.Default: ActionE.Reject,
    },
    {
        # Follow to loopback, but alternately reject/return others, flipped from the previous scenario.
        AddressE.Private: ActionE.Return,
        AddressE.Loopback: ActionE.Follow,
        AddressE.Multicast: ActionE.Return,
        AddressE.Linklocal: ActionE.Reject,
        AddressE.Routable: ActionE.Return,
        AddressE.Self: ActionE.Reject,
        AddressE.Default: ActionE.Return,
    },
    {
        # Return loopback, but reject everything else.
        AddressE.Loopback: ActionE.Return,
        AddressE.Default: ActionE.Reject,
    },
    {
        # Reject loopback, but return everything else.
        AddressE.Loopback: ActionE.Reject,
        AddressE.Default: ActionE.Return,
    },
    {
        # Return everything.
        AddressE.Default: ActionE.Return,
    },
]

for scenario in scenarios:
    for addressClass in AddressE:
        if not addressClass.value:
            # Default has no particular addresses to test.
            continue
        for address in addressClass.value:
            expectedAction = scenario[addressClass] if addressClass in scenario else scenario[AddressE.Default]
            makeTestCase(redirectTarget=address, expectedAction=expectedAction, scenario=scenario)

    # Test redirects to names that cannot be resolved.
    makeTestCase(redirectTarget=None, expectedAction=ActionE.Break, scenario=scenario)

dns.addRecords(records=dnsRecords)

# Make sure this runs only after local files have been created.
Test.Setup.Copy(data_path)



'''
Verify correct statichit plugin behavior
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Verify correct statichit plugin behavior
'''

Test.SkipUnless(Condition.PluginExists('statichit.so'),)


class RemapData:
    """Data for each remap rule"""

    @staticmethod
    def setup():
        # Path relative to config directory.  The statichit directory will be created, and small_body.txt copied to it,
        # by the 2nd (dummy) test run.
        RemapData.small_body_path = os.path.join('statichit', 'small_body.txt')

        Test.Setup.Copy('story_16.json', Test.RunDirectory)
        RemapData.json_body_path = os.path.join(Test.RunDirectory, 'story_16.json')

        Test.Setup.Copy('empty.txt', Test.RunDirectory)
        RemapData.empty_body_path = os.path.join(Test.RunDirectory, 'empty.txt')

        # This directory will be created, and a file copied to it, by the 1st (dummy) test run.
        RemapData.dir_path = os.path.join(Test.RunDirectory, 'body_dir')

    def __init__(self, from_url, *args) -> None:
        self.remap_from = from_url
        self.plugin_args = args
        self.body = RemapData.small_body_path

    def bad_body(self):
        self.body = 'invalid'
        return self

    def json_body(self):
        self.body = RemapData.json_body_path
        return self

    def empty_body(self):
        self.body = RemapData.empty_body_path
        return self

    def dir(self):
        self.body = RemapData.dir_path
        return self


RemapData.setup()

remap_data = [
    RemapData('http://fqdn1'),
    RemapData('http://fqdn2', '--max-age=123'),
    RemapData('http://fqdn3', '--success-code=200', '--disable-exact', '--failure-code=222'),
    RemapData('http://fqdn4', '--success-code=200', '--failure-code=412').bad_body(),
    RemapData('http://fqdn5', '--success-code=200', '--failure-code=412', "--mime-type=application/json").json_body(),
    RemapData('http://fqdn6', '--success-code=200', '--failure-code=222').empty_body(),
    RemapData('http://fqdn7', '--success-code', '200', '--disable-exact', '--failure-code', '222'),
    RemapData('http://fqdn8', '-s', '200', '-d', '-c', '222'),
    RemapData('http://fqdn9').dir(),
]

ts = Test.MakeATSProcess('ts')

ts.Disk.records_config.update({
    "proxy.config.diags.debug.enabled": 1,
    "proxy.config.diags.debug.tags": "http|statichit",
})

for d in remap_data:
    arg_str = ''
    for arg in d.plugin_args:
        arg_str += ' @pparam=' + arg
    ts.Disk.remap_config.AddLine(f'map {d.remap_from} http://127.0.0.1/ @plugin=statichit.so @pparam=--file-path={d.body}{arg_str}')

# Dummy test run for copying a file.
tr = Test.AddTestRun()
p = tr.Processes.Default
p.Command = (f'mkdir {Test.RunDirectory}/body_dir ; ' + f'cp {Test.TestDirectory}/small_body.txt {Test.RunDirectory}/body_dir/.')

# Dummy test run for copying a file.
tr = Test.AddTestRun()
p = tr.Processes.Default
p.Command = (
    f'mkdir {ts.Variables.CONFIGDIR}/statichit ; ' + f'cp {Test.TestDirectory}/small_body.txt {ts.Variables.CONFIGDIR}/statichit/.')
p.StartBefore(ts)
p.StillRunningAfter = ts

tr = Test.AddTestRun()
p = tr.AddVerifierClientProcess('client', 'statichit.replay.yaml', http_ports=[ts.Variables.port], other_args='--thread-limit 1')
p.StillRunningAfter = ts



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test a basic null transform plugin
'''

Test.SkipUnless(Condition.PluginExists('null_transform.so'))

Test.ContinueOnFail = True

# Define default ATS
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

Test.testName = ""
request_header = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# Expected response from origin server
response_header = {
    "headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n",
    "timestamp": "1469733493.993",
    "body": "This is expected response."
}

# Add response the server dictionary
server.addResponse("sessionfile.log", request_header, response_header)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.show_location': 0,
        'proxy.config.diags.debug.tags': 'null_transform'
    })
ts.Disk.remap_config.AddLine('map http://www.example.com http://127.0.0.1:{0}'.format(server.Variables.Port))

# Load plugin
Test.PrepareInstalledPlugin('null_transform.so', ts)

# www.example.com Host
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --proxy 127.0.0.1:{0} "http://www.example.com" --verbose -H "Proxy-Connection: keep-alive"'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stderr = "gold/null_transform-200.gold"
tr.StillRunningAfter = server

# Check Plugin Loading Information
ts.Disk.traffic_out.Content = "gold/null_transform-tag.gold"



'''
Test the client_context_dump plugin.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test client_context_dump plugin
'''

Test.SkipUnless(Condition.PluginExists('client_context_dump.so'))

# Set up ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)

# Set up ssl files
ts.addSSLfile("ssl/one.com.pem")
ts.addSSLfile("ssl/two.com.pem")

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'client_context_dump',
        'proxy.config.ssl.server.cert.path': '{}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.cert.path': '{}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.client.private_key.path': '{}'.format(ts.Variables.SSLDir),
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=one.com.pem ssl_key_name=one.com.pem')

ts.Disk.sni_yaml.AddLines(
    [
        'sni:',
        '- fqdn: "*one.com"',
        '  client_cert: "one.com.pem"',
        '- fqdn: "*two.com"',
        '  client_cert: "two.com.pem"',
    ])

# Set up plugin
Test.PrepareInstalledPlugin('client_context_dump.so', ts)

# custom log comparison.  Verify the two certs we have loaded are dumped
log = Test.Disk.File(ts.Variables.LOGDIR + '/client_context_dump.log', exists=True)
log.Content = Testers.ContainsExpression('CN=two.com', 'Info on two.com.pem should dump')
log.Content += Testers.ContainsExpression("CN=one.com", "Info on one.com.pem should dump")

# traffic server test
t = Test.AddTestRun("Test traffic server started properly")
t.StillRunningAfter = Test.Processes.ts

p = t.Processes.Default
p.Command = "curl http://127.0.0.1:{0}".format(ts.Variables.port)
p.ReturnCode = 0
p.StartBefore(Test.Processes.ts)

# Client contexts test
tr = Test.AddTestRun()
tr.Processes.Default.Env = ts.Env
tr.Processes.Default.Command = ('{0}/traffic_ctl plugin msg client_context_dump.t 1'.format(ts.Variables.BINDIR))
tr.Processes.Default.ReturnCode = 0



"""
Verify traffic_dump functionality.
"""
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

Test.Summary = '''
Verify traffic_dump functionality.
'''

Test.SkipUnless(Condition.PluginExists('traffic_dump.so'),)

schema_path = os.path.join(Test.Variables.AtsTestToolsDir, 'lib', 'replay_schema.json')

# Configure the origin server.
replay_file = "replay/traffic_dump.yaml"
server = Test.MakeVerifierServerProcess("server", replay_file, ssl_cert="ssl/server_combined.pem", ca_cert="ssl/signer.pem")

# Define ATS and configure it.
ts = Test.MakeATSProcess("ts", enable_tls=True)
replay_dir = os.path.join(ts.RunDirectory, "ts", "log")

ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")
ts.addSSLfile("ssl/signer.pem")

ts.Setup.Copy("ssl/signed-foo.pem")
ts.Setup.Copy("ssl/signed-foo.key")

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'traffic_dump|http',
        'proxy.config.http.insert_age_in_response': 0,
        'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.ssl.CA.cert.filename': f'{ts.Variables.SSLDir}/signer.pem',
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.http.host_sni_policy': 2,
        'proxy.config.ssl.TLSv1_3.enabled': 0,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
        'proxy.config.http.connect_ports': f"{server.Variables.http_port}",
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLines(
    [
        f'map https://www.client_only_tls.com/ http://127.0.0.1:{server.Variables.http_port}',
        f'map https://www.tls.com/ https://127.0.0.1:{server.Variables.https_port}',
        f'map http://www.connect_target.com/ http://127.0.0.1:{server.Variables.http_port}',
        f'map / http://127.0.0.1:{server.Variables.http_port}',
    ])

# Configure traffic_dump.
ts.Disk.plugin_config.AddLine(
    f'traffic_dump.so --logdir {replay_dir} --sample 1 --limit 1000000000 '
    '--sensitive-fields "cookie,set-cookie,x-request-1,x-request-2"')
# Configure logging of transactions. This is helpful for the cache test below.
ts.Disk.logging_yaml.AddLines(
    '''
logging:
  formats:
    - name: basic
      format: "%<cluc>: Read result: %<crc>:%<crsc>:%<chm>, Write result: %<cwr>"
  logs:
    - filename: transactions
      format: basic
'''.split('\n'))

# Set up trafficserver expectations.
ts.Disk.diags_log.Content = Testers.ContainsExpression(
    "loading plugin.*traffic_dump.so", "Verify the traffic_dump plugin got loaded.")
ts.Disk.traffic_out.Content = Testers.ContainsExpression(
    f"Initialized with log directory: {replay_dir}", "Verify traffic_dump initialized with the configured directory.")
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Initialized with sample pool size of 1 bytes and disk limit of 1000000000 bytes",
    "Verify traffic_dump initialized with the configured disk limit.")
ts.Disk.traffic_out.Content += Testers.ContainsExpression(
    "Finish a session with log file of.*bytes", "Verify traffic_dump sees the end of sessions and accounts for it.")
ts.Disk.traffic_out.Content += Testers.ContainsExpression("Dumping body bytes: false", "Verify that dumping body bytes is enabled.")

# Set up the json replay file expectations.
replay_file_session_1 = os.path.join(replay_dir, "127", "0000000000000000")
ts.Disk.File(replay_file_session_1, exists=True)
replay_file_session_2 = os.path.join(replay_dir, "127", "0000000000000001")
ts.Disk.File(replay_file_session_2, exists=True)
replay_file_session_3 = os.path.join(replay_dir, "127", "0000000000000002")
ts.Disk.File(replay_file_session_3, exists=True)
replay_file_session_4 = os.path.join(replay_dir, "127", "0000000000000003")
ts.Disk.File(replay_file_session_4, exists=True)
replay_file_session_5 = os.path.join(replay_dir, "127", "0000000000000004")
ts.Disk.File(replay_file_session_5, exists=True)
replay_file_session_6 = os.path.join(replay_dir, "127", "0000000000000005")
ts.Disk.File(replay_file_session_6, exists=True)
replay_file_session_7 = os.path.join(replay_dir, "127", "0000000000000006")
ts.Disk.File(replay_file_session_7, exists=True)
replay_file_session_8 = os.path.join(replay_dir, "127", "0000000000000007")
ts.Disk.File(replay_file_session_8, exists=True)
replay_file_session_9 = os.path.join(replay_dir, "127", "0000000000000008")
ts.Disk.File(replay_file_session_9, exists=True)
replay_file_session_10 = os.path.join(replay_dir, "127", "0000000000000009")
ts.Disk.File(replay_file_session_10, exists=True)
replay_file_session_11 = os.path.join(replay_dir, "127", "000000000000000a")
ts.Disk.File(replay_file_session_11, exists=True)

# The following will not be written to disk because of a restricted disk limit.
replay_file_session_12 = os.path.join(replay_dir, "127", "000000000000000b")
ts.Disk.File(replay_file_session_12, exists=False)

# The following will be written to disk because the disk restriction will be
# removed.
replay_file_session_13 = os.path.join(replay_dir, "127", "000000000000000c")
ts.Disk.File(replay_file_session_13, exists=True)

# The following will not be written to disk because the restriction will be
# re-added.
replay_file_session_14 = os.path.join(replay_dir, "127", "000000000000000d")
ts.Disk.File(replay_file_session_14, exists=False)

# Run our test traffic.
tr = Test.AddTestRun("Run the test traffic.")
tr.AddVerifierClientProcess(
    "client",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--thread-limit 1')

tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 1: Verify the correct behavior of two transactions across two sessions.
#

# Verify the properties of the replay file for the first transaction.
tr = Test.AddTestRun("Verify the json content of the first session")
http_protocols = "tcp,ip"
verify_replay = "verify_replay.py"
sensitive_fields_arg = (
    "--sensitive-fields cookie "
    "--sensitive-fields set-cookie "
    "--sensitive-fields x-request-1 "
    "--sensitive-fields x-request-2 ")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_1} '
     f'{sensitive_fields_arg} --client-http-version "1.1" '
     f'--client-protocols "{http_protocols}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

# Verify the properties of the replay file for the second transaction.
tr = Test.AddTestRun("Verify the json content of the second session")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_2} '
     f'{sensitive_fields_arg} --client-http-version "1.1" '
     '--request-target "/two"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 2: Verify the correct behavior of an explicit path in the request line.
#

# Verify recording of a request target with the host specified.
tr = Test.AddTestRun("Verify the replay file has the explicit target.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)

tr.Processes.Default.Command = \
    (f"{sys.executable} {verify_replay} {schema_path} {replay_file_session_3} {sensitive_fields_arg} "
     "--request-target 'http://www.some.host.com/candy'")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 3: Verify correct handling of a POST with body data.
#

tr = Test.AddTestRun("Verify the client-request size node for a request with a body.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)

size_of_verify_replay_file = os.path.getsize(os.path.join(Test.TestDirectory, verify_replay))
expected_body_size = 12345
tr.Processes.Default.Command = \
    (f"{sys.executable} {verify_replay} {schema_path} {replay_file_session_4} {sensitive_fields_arg} "
     f"--client-request-size {expected_body_size}")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 4: Verify correct handling of a response produced out of the cache.
#
tr = Test.AddTestRun("Verify that the cached response's replay file looks appropriate.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_6} '
     f'--client-protocols "{http_protocols}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 5: Verify correct handling of two transactions in a session.
#
tr = Test.AddTestRun("Verify the dump file of two transactions in a session.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_7} '
     f'--client-protocols "{http_protocols}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 6: Verify correct protocol dumping of a TLS connection.
#
tr = Test.AddTestRun("Verify the client protocol stack of a TLS session.")
https_protocols = "tls,tcp,ip"
client_tls_features = "sni:www.tls.com,proxy-verify-mode:0,proxy-provided-cert:true"
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_8} '
     f'--client-protocols "{https_protocols}" --client-tls-features "{client_tls_features}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

tr = Test.AddTestRun("Verify the server TLS protocol stack.")
https_server_stack = "http,tls,tcp,ip"
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
server_tls_features = 'proxy-provided-cert:false,sni:www.tls.com,proxy-verify-mode:1'
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_8} --server-protocols '
     f'"{https_server_stack}" --server-tls-features "{server_tls_features}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 7: Verify correct protocol dumping of TLS and HTTP/2 connections.
#
tr = Test.AddTestRun("Verify the client HTTP/2 protocol stack.")
h2_protocols = "http,tls,tcp,ip"
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_9} '
     f'--client-http-version "2" --client-protocols "{h2_protocols}" '
     f'--client-tls-features "{client_tls_features}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

tr = Test.AddTestRun("Verify the server HTTP/2 protocol stack.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_9} '
     f'--server-protocols "{https_server_stack}" --server-tls-features "{server_tls_features}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 8: Verify correct protocol dumping of client-side TLS and server-side HTTP.
#
tr = Test.AddTestRun("Verify the client TLS protocol stack.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_10} '
     f'--client-http-version "1.1" --client-protocols "{https_protocols}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

tr = Test.AddTestRun("Verify the server HTTP protocol stack.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
http_server_stack = "http,tcp,ip"
tr.Processes.Default.Command = \
    (f'{sys.executable} {verify_replay} {schema_path} {replay_file_session_10} '
     f'--server-protocols "{http_server_stack}"')
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 9: Verify correct handling of a CONNECT request.
#

tr = Test.AddTestRun("Verify handling of a CONNECT request.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)

tr.Processes.Default.Command = \
    (f"{sys.executable} {verify_replay} {schema_path} {replay_file_session_11} {sensitive_fields_arg} ")
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Test 10: Verify that we can change the --limit value.
#
tr = Test.AddTestRun("Verify changing --limit via traffic_ctl.")
tr.Processes.Default.Command = "traffic_ctl plugin msg traffic_dump.limit 0"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts

# Give ATS some time to process the change
tr = Test.AddTestRun("Give ATS some time to process the change")
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("Run some more test traffic with the restricted disk limit.")
tr.AddVerifierClientProcess(
    "client-2",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--keys 1')

# Since the limit is zero, we should not see any new replay file created.
tr = Test.AddTestRun("Verify no new traffic was dumped")
# Sleep 2 seconds to give the replay plugin plenty of time to write the file.
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0
file = tr.Disk.File(replay_file_session_12)
file.Exists = False

#
# Test 11: Verify that we can remove the disk limit.
#
tr = Test.AddTestRun("Removing the disk limit via traffic_ctl.")
tr.Processes.Default.Command = "traffic_ctl plugin msg traffic_dump.unlimit"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts

# Give ATS some time to process the change
tr = Test.AddTestRun("Give ATS some time to process the change")
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("Run some more test traffic with no disk limit.")
tr.AddVerifierClientProcess(
    "client-3",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--keys 1')

# Since the limit is zero, we should not see any new replay file created.
tr = Test.AddTestRun("Verify the new traffic was dumped")
# Sleep 2 seconds to give the replay plugin plenty of time to write the file.
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0
file = tr.Disk.File(replay_file_session_13)
file.Exists = True

#
# Test 11: Verify that we can again restrict the disk limit.
#
# Verify that the restriction can be re-added after unlimit was set. This
# verifies correct handling of the boolean controlling unlimited disk space.
#
tr = Test.AddTestRun("Verify re-adding --limit via traffic_ctl.")
tr.Processes.Default.Command = "traffic_ctl plugin msg traffic_dump.limit 0"
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Env = ts.Env
tr.StillRunningAfter = ts

# Give ATS some time to process the change
tr = Test.AddTestRun("Give ATS some time to process the change")
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun("Run test traffic with newly restricted disk limit.")
tr.AddVerifierClientProcess(
    "client-4",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--keys 1')

# Since the limit is zero, we should not see any new replay file created.
tr = Test.AddTestRun("Verify no new traffic was dumped")
# Sleep 2 seconds to give the replay plugin plenty of time to write the file.
tr.Processes.Default.Command = "sleep 2"
tr.Processes.Default.ReturnCode = 0
file = tr.Disk.File(replay_file_session_14)
file.Exists = False



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test header_rewrite with METHOD conditions and operators.
'''

Test.ContinueOnFail = True
# Define default ATS
ts = Test.MakeATSProcess("ts")
server = Test.MakeOriginServer("server")

Test.testName = "header_rewrite_method_condition"
request_get = {"headers": "GET / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
request_delete = {"headers": "DELETE / HTTP/1.1\r\nHost: www.example.com\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
# expected response from the origin server
response = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}

# add response to the server dictionary
session_file = "sessionfile.log"
server.addResponse(session_file, request_get, response)
server.addResponse(session_file, request_delete, response)
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.show_location': 0,
        'proxy.config.diags.debug.tags': 'header.*',
        'proxy.config.http.insert_response_via_str': 0,
    })
# The following rule inserts a via header if the request method is a GET or DELETE
conf_name = "rule_cond_method.conf"
ts.Setup.CopyAs('rules/{0}'.format(conf_name), Test.RunDirectory)
ts.Disk.plugin_config.AddLine('header_rewrite.so {0}/{1}'.format(Test.RunDirectory, conf_name))
ts.Disk.remap_config.AddLine('map http://www.example.com http://127.0.0.1:{0}'.format(server.Variables.Port))

# Test method in READ_REQUEST_HDR_HOOK.
expected_output = "gold/header_rewrite_cond_method.gold"
expected_log = "gold/header_rewrite-tag.gold"
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --proxy 127.0.0.1:{0} "http://www.example.com" -H "Proxy-Connection: keep-alive" --verbose'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.StartBefore(server, ready=When.PortOpen(server.Variables.Port))
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.Processes.Default.Streams.stderr = expected_output
tr.StillRunningAfter = server
ts.Disk.traffic_out.Content = expected_log

# Test method in SEND_REQUEST_HDR_HOOK.
tr = Test.AddTestRun()
tr.Processes.Default.Command = 'curl --request DELETE --proxy 127.0.0.1:{0} "http://www.example.com" -H "Proxy-Connection: keep-alive" --verbose'.format(
    ts.Variables.port)
tr.Processes.Default.ReturnCode = 0
tr.Processes.Default.Streams.stderr = expected_output
tr.StillRunningAfter = server
ts.Disk.traffic_out.Content = expected_log



'''
Verify h3 SNI checking behavior.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Verify h3 SNI checking behavior.
'''

Test.SkipUnless(Condition.HasATSFeature('TS_HAS_QUICHE'), Condition.HasCurlFeature('http3'))

Test.ContinueOnFail = True


class Test_sni_check:
    """Configure a test to verify SNI checking behavior for h3 connections."""

    replay_file = "replays/h3_sni.replay.yaml"
    client_counter: int = 0
    ts_counter: int = 0
    server_counter: int = 0

    def __init__(self, name: str, gold_file="", replay_keys="", expect_request_rejected=False):
        """Initialize the test.

        :param name: The name of the test.
        :param gold_file: (Optional) Gold file to be checked.
        :param replay_keys: (Optional) Keys to be used by pv.
        :param expect_request_rejected: (Optional) Whether or not the client request is expected to be rejected.
        """
        self.name = name
        self.gold_file = gold_file
        self.replay_keys = replay_keys
        self.expect_request_rejected = expect_request_rejected

    def _configure_server(self, tr: 'TestRun'):
        """Configure the server.

        :param tr: The TestRun object to associate the server process with.
        """
        server = tr.AddVerifierServerProcess(f"server_{Test_sni_check.server_counter}", self.replay_file)
        Test_sni_check.server_counter += 1
        self._server = server

    def _configure_traffic_server(self, tr: 'TestRun'):
        """Configure Traffic Server.

        :param tr: The TestRun object to associate the ts process with.
        """
        ts = tr.MakeATSProcess(f"ts-{Test_sni_check.ts_counter}", enable_quic=True, enable_tls=True)

        Test_sni_check.ts_counter += 1
        self._ts = ts
        # Configure TLS for Traffic Server.
        self._ts.addDefaultSSLFiles()
        self._ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')
        self._ts.Disk.records_config.update(
            {
                'proxy.config.diags.debug.enabled': 1,
                'proxy.config.diags.debug.tags': 'http',
                'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
                'proxy.config.quic.no_activity_timeout_in': 0,
                'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
            })

        self._ts.Disk.remap_config.AddLine(f'map / http://127.0.0.1:{self._server.Variables.http_port}')

    def run(self):
        """Run the test."""
        tr = Test.AddTestRun(self.name)
        self._configure_server(tr)
        self._configure_traffic_server(tr)

        tr.Processes.Default.StartBefore(self._server)
        tr.Processes.Default.StartBefore(self._ts)

        tr.AddVerifierClientProcess(
            f'client-{Test_sni_check.client_counter}',
            self.replay_file,
            http3_ports=[self._ts.Variables.ssl_port],
            keys=self.replay_keys)
        Test_sni_check.client_counter += 1

        if self.expect_request_rejected:
            # The client request should time out because ATS rejects it and does
            # not send a response.
            tr.Processes.Default.ReturnCode = 1
            self._ts.Disk.traffic_out.Content += Testers.ContainsExpression("SNI not found", "ATS should detect the missing SNI.")
        else:
            # Verify the client request is successful.
            tr.Processes.Default.ReturnCode = 0
            self._ts.Disk.traffic_out.Content += Testers.ExcludesExpression(
                "SNI not found", "ATS should see the SNI presented by client.")

        if self.gold_file:
            tr.Processes.Default.Streams.all = self.gold_file


# TEST 1: Client request with SNI.
test0 = Test_sni_check("", replay_keys="has_sni", expect_request_rejected=False)
test0.run()

# TEST 2: Client request without SNI.
test1 = Test_sni_check("", replay_keys="no_sni", expect_request_rejected=True)
test1.run()



"""
Verify traffic_dump response body functionality.
"""
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

Test.Summary = '''
Verify traffic_dump response body functionality.
'''

Test.SkipUnless(Condition.PluginExists('traffic_dump.so'),)

# Configure the origin server.
replay_file = "replay/response_body.yaml"
server = Test.MakeVerifierServerProcess("server", replay_file)

# Configure ATS.
ts = Test.MakeATSProcess("ts", enable_tls=True)
replay_dir = os.path.join(ts.RunDirectory, "ts", "log")

ts.addSSLfile("ssl/server.pem")
ts.addSSLfile("ssl/server.key")
ts.addSSLfile("ssl/signer.pem")

ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'traffic_dump',
        'proxy.config.ssl.server.cert.path': ts.Variables.SSLDir,
        'proxy.config.ssl.server.private_key.path': ts.Variables.SSLDir,
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.ssl.CA.cert.filename': f'{ts.Variables.SSLDir}/signer.pem',
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.http.host_sni_policy': 2,
        'proxy.config.ssl.TLSv1_3.enabled': 0,
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

ts.Disk.remap_config.AddLines([
    f'map / http://127.0.0.1:{server.Variables.http_port}',
])

# Configure traffic_dump to dump body bytes (-b).
ts.Disk.plugin_config.AddLine(f'traffic_dump.so --logdir {replay_dir} --sample 1 --limit 1000000000 -b')

ts_dump_0 = os.path.join(replay_dir, "127", "0000000000000000")
ts.Disk.File(ts_dump_0, exists=True)

ts_dump_1 = os.path.join(replay_dir, "127", "0000000000000001")
ts.Disk.File(ts_dump_1, exists=True)

ts_dump_2 = os.path.join(replay_dir, "127", "0000000000000002")
ts.Disk.File(ts_dump_2, exists=True)

ts_dump_3 = os.path.join(replay_dir, "127", "0000000000000003")
ts.Disk.File(ts_dump_3, exists=True)

ts.Disk.traffic_out.Content = Testers.ContainsExpression("Dumping body bytes: true", "Verify that dumping body bytes is enabled.")

# Run our test traffic.
tr = Test.AddTestRun("Run the test traffic.")
tr.AddVerifierClientProcess(
    "client",
    replay_file,
    http_ports=[ts.Variables.port],
    https_ports=[ts.Variables.ssl_port],
    ssl_cert="ssl/server_combined.pem",
    ca_cert="ssl/signer.pem",
    other_args='--thread-limit 1')

tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

# Common verification variables.
verify_replay = "verify_replay.py"
schema = os.path.join(Test.Variables.AtsTestToolsDir, 'lib', 'replay_schema.json')
verify_command_prefix = f'{sys.executable} {verify_replay} {schema}'

#
# Verify a response without a body is dumped correctly.
#
tr = Test.AddTestRun("Verify the json content of the transaction with no response body.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = f'{verify_command_prefix} {ts_dump_0}'
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Verify a response with a body is dumped correctly.
#
tr = Test.AddTestRun("Verify the json content of the transaction with a response body.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    f'{verify_command_prefix} {ts_dump_1} --response_body "0000000 0000001 "'
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts

#
# Verify a response with a character to be escaped is dumped correctly.
#
tr = Test.AddTestRun("Verify the json content of the response body with a character to be escaped.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    rf'{verify_command_prefix} {ts_dump_2} --response_body 12\"34'
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
#
# Verify an HTTP/2 response with a body is dumped correctly.
#
tr = Test.AddTestRun("Verify the json content of the response body of an HTTP/2 transaction.")
tr.Setup.CopyAs(verify_replay, Test.RunDirectory)
tr.Processes.Default.Command = \
    f'{verify_command_prefix} {ts_dump_3} --response_body "0000000 0000001 0000002 "'
tr.Processes.Default.ReturnCode = 0
tr.StillRunningAfter = server
tr.StillRunningAfter = ts



#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test TS API.
'''

Test.SkipUnless(Condition.HasCurlFeature('http2'),)
Test.ContinueOnFail = True

plugin_name = "test_tsapi"

server = Test.MakeOriginServer("server")

request_header = {"headers": "GET / HTTP/1.1\r\nHost: doesnotmatter\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": "112233"}
server.addResponse("sessionlog.json", request_header, response_header)

request_header = {"headers": "GET /xYz HTTP/1.1\r\nHost: doesnotmatter\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": "445566"}
server.addResponse("sessionlog.json", request_header, response_header)

# Disable the cache to make sure each request is forwarded to the origin
# server.
ts = Test.MakeATSProcess("ts", enable_tls=True, enable_cache=False)

# The test plugin will output test logging to this file.
log_file_name = os.path.join(ts.Variables.LOGDIR, "log.txt")
Test.Env["OUTPUT_FILE"] = log_file_name

ts.addDefaultSSLFiles()

ts.Disk.records_config.update(
    {
        'proxy.config.proxy_name': 'Poxy_Proxy',  # This will be the server name.
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.url_remap.remap_required': 1,
        'proxy.config.diags.debug.enabled': 3,
        'proxy.config.diags.debug.tags': f'http|{plugin_name}',
    })

ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

rp = os.path.join(Test.Variables.AtsBuildGoldTestsDir, 'pluginTest', 'tsapi', '.libs', f'{plugin_name}.so')
ts.Setup.Copy(rp, ts.Env['PROXY_CONFIG_PLUGIN_PLUGIN_DIR'])

ts.Disk.remap_config.AddLine(
    "map http://myhost.test http://127.0.0.1:{0} @plugin={1} @plugin={1}".format(server.Variables.Port, f"{plugin_name}.so"))
ts.Disk.remap_config.AddLine(
    "map https://myhost.test:123 http://127.0.0.1:{0} @plugin={1} @plugin={1}".format(server.Variables.Port, f"{plugin_name}.so"))

# For some reason, without this delay, traffic_server cannot reliably open the cleartext port for listening without an
# error.
#
tr = Test.AddTestRun()
tr.Processes.Default.Command = "sleep 3"
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
#
tr.Processes.Default.Command = ('curl --verbose --ipv4 --header "Host: mYhOsT.teSt" hTtP://loCalhOst:{}/'.format(ts.Variables.port))
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = ('curl --verbose --ipv4 --proxy localhost:{} http://mYhOsT.teSt/xYz'.format(ts.Variables.port))
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
tr.Processes.Default.Command = (
    'curl --verbose --ipv4 --http2 --insecure --header ' +
    '"Host: myhost.test:123" HttPs://LocalHost:{}/'.format(ts.Variables.ssl_port))
tr.Processes.Default.ReturnCode = 0

tr = Test.AddTestRun()
# Change server port number (which can vary) to a fixed string for compare to gold file.
second_log_file_name = os.path.join(ts.Variables.LOGDIR, "log2.txt")
tr.Processes.Default.Command = f"sed 's/{server.Variables.Port}/SERVER_PORT/' < {log_file_name} > {second_log_file_name}"
tr.Processes.Default.ReturnCode = 0
f = tr.Disk.File(second_log_file_name)
f.Content = "log.gold"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
##
# http://www.apache.org/licenses/LICENSE-2.0
##
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# This script will add a command to help print out ATS data structures in gdb.
#
# In a .gdbinit file or from the gdb command prompt:
#   source <path to this file>.py
#
# Then you can type:  `atspr sm this`  Where this is a HttpSM* and it will print some information about it.
#
#  For the expression you should be able to put anything that can be cast to the appropriate type.
#
# Type `atspr` by itself for a list of printables.
#

import gdb
import socket
import struct
from curses.ascii import isgraph

# return memory for an ats string address and length


def ats_str(addr, addr_len):
    if addr_len == 0:
        return ''
    #print("addr {} len {}".format(addr, addr_len))
    inferior = gdb.selected_inferior()
    try:
        buff = inferior.read_memory(addr, addr_len)
        if buff[0] == '\x00':
            return 'null'
        else:
            return buff
    except BaseException:
        return 'unreadable({:x})'.format(int(addr))


def hdrtoken(idx):
    hdrtokens = gdb.parse_and_eval('hdrtoken_strs')
    return hdrtokens[idx].string()


def wks_or_str(idx, addr, addr_len):
    if idx >= 0:
        return hdrtoken(idx)
    else:
        return ats_str(addr, addr_len)


class URL:

    def __init__(self, val):
        self.impl = val['m_url_impl'].dereference()

    def __str__(self):
        return "{}://{}/{}".format(self.scheme(), self.host(), self.path())

    def scheme(self):
        return ats_str(self.impl['m_ptr_scheme'], self.impl['m_len_scheme'])

    def host(self):
        return ats_str(self.impl['m_ptr_host'], self.impl['m_len_host'])

    def path(self):
        return ats_str(self.impl['m_ptr_path'], self.impl['m_len_path'])


class HTTPHdr:

    def __init__(self, val):
        self.impl = val['m_http']
        self.val = val

    def url(self):
        return URL(self.val['m_url_cached'])

    def is_request(self):
        pol = self.impl['m_polarity']
        return pol == 1

    def is_response(self):
        pol = self.impl['m_polarity']
        return pol == 2

    def is_valid(self):
        if self.impl == 0:
            return None
        pol = self.impl['m_polarity']
        return pol > 0

    def method(self):
        idx = self.impl['u']['req']['m_method_wks_idx']
        if idx >= 0:
            return hdrtoken(idx)
        else:
            return ats_str(self.impl['u']['req']['m_ptr_method'], self.impl['u']['req']['m_len_method'])

    def status(self):
        return self.impl['u']['resp']['m_status']

    def headers(self):
        mime = self.val['m_mime'].dereference()
        fblock_ptr = mime['m_first_fblock'].address
        while fblock_ptr != 0:
            fblock = fblock_ptr.dereference()
            slots = fblock['m_field_slots']
            #print("slots type {} address {} size {}".format(slots.type, slots.address, slots.type.sizeof))
            #print("next idx {} len {} next {}".format(fblock['m_freetop'], fblock['m_length'], fblock['m_next']))

            for slot_idx in range(fblock['m_freetop']):
                fld = slots[slot_idx]
                wks = fld['m_wks_idx']
                name = wks_or_str(wks, fld['m_ptr_name'], fld['m_len_name'])
                yield (name, ats_str(fld['m_ptr_value'], fld['m_len_value']))

            fblock_ptr = fblock['m_next']

    def pr(self):
        if self.is_valid():
            if self.is_request():
                print("{} {}".format(self.method(), self.url()))
            if self.is_response():
                print("status: {}".format(self.status()))
            for key, val in self.headers():
                print("{}: {}".format(key, val))
        else:
            print("invalid")


class ConnectionAttributes:

    def __init__(self, val):
        self.val = val

    def http_version(self):
        v = self.val['http_version']['m_version']
        return (v >> 16, v & 0xffff)

    def src_addr(self):
        ip = self.val['src_addr']
        s = struct.pack('!I', socket.ntohl(int(ip['sin']['sin_addr']['s_addr'])))
        return socket.inet_ntoa(s)

    def dst_addr(self):
        ip = self.val['dst_addr']
        s = struct.pack('!I', socket.ntohl(int(ip['sin']['sin_addr']['s_addr'])))
        return socket.inet_ntoa(s)


class HttpSM:

    def __init__(self, val):
        ptr_type = gdb.lookup_type('struct HttpSM').pointer()
        self.val = None

        if val.type != ptr_type:
            self.val = val.cast(ptr_type)
        else:
            self.val = val

    def id(self):
        return self.val['sm_id']

    def client_request(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['client_request'])

    def client_response(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['client_response'])

    def server_request(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['server_request'])

    def server_response(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['server_response'])

    def transform_response(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['transform_response'])

    def cache_response(self):
        return HTTPHdr(self.val['t_state']['hdr_info']['cache_response'])

    def client_info(self):
        return ConnectionAttributes(self.val['t_state']['client_info'])

    def parent_info(self):
        return ConnectionAttributes(self.val['t_state']['client_info'])

    def server_info(self):
        return ConnectionAttributes(self.val['t_state']['client_info'])


def sm_command(val):
    sm = HttpSM(val)

    print('smid = %s' % sm.id())
    print('---- request ----')
    sm.client_request().pr()
    print('---- response ----')
    sm.client_response().pr()
    print('---- server_request  ----')
    sm.server_request().pr()
    print('---- server_response  ----')
    sm.server_response().pr()
    print('---- transform_response  ----')
    sm.transform_response().pr()
    print('---- cache_response  ----')
    sm.cache_response().pr()
    print('---- client_info  ----')
    info = sm.client_info()
    major, minor = info.http_version()
    print("HTTP {}.{} ({} -> {})".format(major, minor, info.src_addr(), info.dst_addr()))
    print('---- server_info  ----')
    info = sm.server_info()
    major, minor = info.http_version()
    print("HTTP {}.{} ({} -> {})".format(major, minor, info.src_addr(), info.dst_addr()))
    print('---- parent_info  ----')
    info = sm.parent_info()
    major, minor = info.http_version()
    print("HTTP {}.{} ({} -> {})".format(major, minor, info.src_addr(), info.dst_addr()))


def hdrs_command(val):
    HTTPHdr(val).pr()


def url_command(val):
    print(URL(val))


commands = [
    ("sm", sm_command, "Print HttpSM details (type HttpSM*)"),
    ("hdrs", hdrs_command, "Print HttpHdr details (type HTTPHdr)"),
    ("url", url_command, "Print URL"),
]


def usage():
    print("Usage: atspr <command> <expr>")
    print("commands:")
    for cmd, f, desc in commands:
        print("  {}: {}".format(cmd, desc))


class ATSPrintCommand(gdb.Command):

    def __init__(self):
        super(ATSPrintCommand, self).__init__('atspr', gdb.COMMAND_DATA)

    def invoke(self, arg, from_tty):
        argv = gdb.string_to_argv(arg)
        if len(argv) < 2:
            usage()
            return

        what = argv[0]
        expr = argv[1]
        val = gdb.parse_and_eval(expr)

        for cmd, f, _ in commands:
            if cmd == what:
                f(val)
                return

        usage()


ATSPrintCommand()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import time

Test.Summary = '''
regex_revalidate plugin test, reload epoch state on ats start
'''

# Test description:
# Ensures that that the regex revalidate config file is loaded,
# then epoch times from the state file are properly merged.

Test.SkipUnless(Condition.PluginExists('regex_revalidate.so'))
Test.ContinueOnFail = False

# configure origin server
server = Test.MakeOriginServer("server")

# Define ATS and configure
ts = Test.MakeATSProcess("ts")

# **testname is required**
testName = "regex_revalidate_state"

# default root
request_header_0 = {
    "headers": "GET / HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "",
}

response_header_0 = {
    "headers": "HTTP/1.1 200 OK\r\n" + "Connection: close\r\n" + "Cache-Control: max-age=300\r\n" + "\r\n",
    "timestamp": "1469733493.993",
    "body": "xxx",
}

server.addResponse("sessionlog.json", request_header_0, response_header_0)

reval_conf_path = os.path.join(ts.Variables.CONFIGDIR, 'reval.conf')
reval_state_path = os.path.join(ts.Variables.RUNTIMEDIR, 'reval.state')

# Configure ATS server
ts.Disk.plugin_config.AddLine(f"regex_revalidate.so -d -c reval.conf -l reval.log -f {reval_state_path}")

sep = ' '

# rule with no initial state
path0_regex = "path0"
path0_expiry = str(time.time() + 90).split('.')[0]
path0_type = "STALE"
path0_rule = sep.join([path0_regex, path0_expiry, path0_type])

path1_regex = "path1"
path1_epoch = str(time.time() - 50).split('.')[0]
path1_expiry = str(time.time() + 600).split('.')[0]
path1_type = "MISS"
path1_rule = sep.join([path1_regex, path1_expiry, path1_type])

# Create gold files
gold_path_good = reval_state_path + ".good"
ts.Disk.File(
    gold_path_good, typename="ats:config").AddLines(
        [
            sep.join([path0_regex, "``", path0_expiry, path0_type]),
            sep.join([path1_regex, path1_epoch, path1_expiry, path1_type]),
        ])

# It seems there's no API for negative gold file matching
'''
gold_path_bad = reval_state_path + ".bad"
ts.Disk.File(gold_path_bad, typename="ats:config").AddLines([
  sep.join([path0_regex, path1_epoch, path0_expiry, path0_type]),
  sep.join([path1_regex, path1_epoch, path1_expiry, path1_type]),
])
'''

# Create a state file, second line will be discarded and not merged
ts.Disk.File(
    reval_state_path, typename="ats:config").AddLines(
        [
            sep.join([path1_regex, path1_epoch, path1_expiry, path1_type]),
            sep.join(["dummy", path1_epoch, path1_expiry, path1_type]),
        ])

# Write out reval.conf file
ts.Disk.File(
    reval_conf_path, typename="ats:config").AddLines([
        path0_rule,
        path1_rule,
    ])

ts.chownForATSProcess(reval_state_path)

ts.Disk.remap_config.AddLine(f"map http://ats/ http://127.0.0.1:{server.Variables.Port}")

# minimal configuration
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'regex_revalidate',
        'proxy.config.http.wait_for_cache': 1,
    })

# This TestRun creates the state file so it exists when the ts process's Setup
# logic is run so that it can be chowned at that point.
tr = Test.AddTestRun("Populate the regex_revalidate state file")
tr.Processes.Default.Command = f'touch {reval_state_path}'

# Start ATS and evaluate the new state file
tr = Test.AddTestRun("Initial load, state merged")
ps = tr.Processes.Default
ps.StartBefore(server, ready=When.PortOpen(server.Variables.Port))

# Note the ready condition: wait for ATS to modify the contents
# of the file from dummy to path1.
ps.StartBefore(Test.Processes.ts, ready=When.FileContains(reval_state_path, "path0"))

ps.Command = 'cat ' + reval_state_path
ps.ReturnCode = 0
ps.Streams.stdout.Content = Testers.GoldFile(gold_path_good)
tr.StillRunningAfter = ts



'''
Test the ESI plugin when origin returns 304 response.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test the ESI plugin when origin returns 304 response.
'''

Test.SkipUnless(Condition.PluginExists('esi.so'),)


class EsiTest():
    """
    A class that encapsulates the configuration and execution of a set of ESI
    test cases.
    """
    """ static: The same server Process is used across all tests. """
    _server = None
    """ static: A counter to keep the ATS process names unique across tests. """
    _ts_counter = 0
    """ static: A counter to keep any output file names unique across tests. """
    _output_counter = 0
    """ The ATS process for this set of test cases. """
    _ts = None

    def __init__(self, plugin_config):
        """
        Args:
            plugin_config (str): The config line to place in plugin.config for
                the ATS process.
        """
        if EsiTest._server is None:
            EsiTest._server = EsiTest._create_server()

        self._ts = EsiTest._create_ats(self, plugin_config)

    @staticmethod
    def _create_server():
        """
        Create and start a server process.
        """
        # Configure our server.
        server = Test.MakeOriginServer("server", lookup_key="{%uuid}")

        # Generate the set of ESI responses.
        request_header = {
            "headers":
                "GET /esi_etag.php HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "uuid: first\r\n" + "Content-Length: 0\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": ""
        }
        esi_body = r'''<html>
<body>
Hello, ESI 304 test
</body>
</html>
'''
        response_header = {
            "headers":
                "HTTP/1.1 200 OK\r\n" + "X-Esi: 1\r\n" + "Cache-Control: public, max-age=0\r\n" + 'Etag: "esi_304_test"\r\n' +
                "Content-Type: text/html\r\n" + "Connection: close\r\n" + "Content-Length: {}\r\n".format(len(esi_body)) + "\r\n",
            "timestamp": "1469733493.993",
            "body": esi_body
        }
        server.addResponse("sessionfile.log", request_header, response_header)

        request_header = {
            "headers":
                "GET /esi_etag.php HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "uuid: second\r\n" +
                'If-None-Match: "esi_304_test"\r\n' + "Content-Length: 0\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": ""
        }
        response_header = {
            "headers":
                "HTTP/1.1 304 Not Modified\r\n" + "Content-Type: text/html\r\n" + "Connection: close\r\n" +
                "Content-Length: 0\r\n" + "\r\n",
            "timestamp": "1469733493.993",
            "body": ""
        }
        server.addResponse("sessionfile.log", request_header, response_header)

        request_header = {
            "headers": "GET /date.php HTTP/1.1\r\n" + "Host: www.example.com\r\n" + "uuid: date\r\n" + "Content-Length: 0\r\n\r\n",
            "timestamp": "1469733493.993",
            "body": ""
        }
        date_body = r'''ESI 304 test
No Date
'''
        response_header = {
            "headers":
                "HTTP/1.1 200 OK\r\n" + "Content-Type: text/html\r\n" + "Connection: close\r\n" +
                "Content-Length: {}\r\n".format(len(date_body)) + "\r\n",
            "timestamp": "1469733493.993",
            "body": date_body
        }
        server.addResponse("sessionfile.log", request_header, response_header)

        # Create a run to start the server.
        tr = Test.AddTestRun("Start the server.")
        tr.Processes.Default.StartBefore(server)
        tr.Processes.Default.Command = "echo starting the server"
        tr.Processes.Default.ReturnCode = 0
        tr.StillRunningAfter = server

        return server

    @staticmethod
    def _create_ats(self, plugin_config):
        """
        Create and start an ATS process.
        """
        EsiTest._ts_counter += 1

        # Configure ATS with a vanilla ESI plugin configuration.
        ts = Test.MakeATSProcess("ts{}".format(EsiTest._ts_counter))
        ts.Disk.records_config.update({
            'proxy.config.diags.debug.enabled': 1,
            'proxy.config.diags.debug.tags': 'http|plugin_esi',
        })
        ts.Disk.remap_config.AddLine('map http://www.example.com/ http://127.0.0.1:{0}'.format(EsiTest._server.Variables.Port))
        ts.Disk.plugin_config.AddLine(plugin_config)

        # Create a run to start the ATS process.
        tr = Test.AddTestRun("Start the ATS process.")
        tr.Processes.Default.StartBefore(ts)
        tr.Processes.Default.Command = "echo starting ATS"
        tr.Processes.Default.ReturnCode = 0
        tr.StillRunningAfter = ts
        return ts

    def run_cases(self):
        # Test 1: Verify basic ESI functionality.
        tr = Test.AddTestRun("First request for esi_etag.php: not cached")
        tr.Processes.Default.Command = \
            ('curl http://127.0.0.1:{0}/esi_etag.php -H"Host: www.example.com" '
             '-H"Accept: */*" -H"uuid: first" --verbose -o /dev/stderr'.format(
                 self._ts.Variables.port))
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.Streams.stderr = "gold/esi_private_headers.gold"
        tr.StillRunningAfter = self._server
        tr.StillRunningAfter = self._ts

        # Test 2: Repeat the above, origin should now be returning 304 response.
        tr = Test.AddTestRun("Second request for esi_etag.php: will be cached")
        tr.Processes.Default.Command = \
            ('curl http://127.0.0.1:{0}/esi_etag.php -H"Host: www.example.com" '
             '-H"Accept: */*" -H"uuid: second" --verbose -o /dev/stderr'.format(
                 self._ts.Variables.port))
        tr.Processes.Default.ReturnCode = 0
        tr.Processes.Default.Streams.stderr = "gold/esi_private_headers.gold"
        tr.StillRunningAfter = self._server
        tr.StillRunningAfter = self._ts


#
# Configure and run the test cases.
#

# Run the tests with ESI configured with private response.
private_response_test = EsiTest(plugin_config='esi.so --private-response')
private_response_test.run_cases()



'''
Test cache matching with the normalized Accept-Encoding header field
and the Vary header field in response
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

Test.Summary = '''
Test cache matching with the normalized Accept-Encoding and the Vary header field in response
'''

Test.ContinueOnFail = True

testName = "NORMALIZE_AE_MATCH_VARY"

replay_file = "replays/normalized_ae_varied_transactions.replay.yaml"
server = Test.MakeVerifierServerProcess("server", replay_file)

ts = Test.MakeATSProcess("ts", enable_cache=True)
ts.Disk.remap_config.AddLine(
    f"map http://www.ae-0.com http://127.0.0.1:{server.Variables.http_port}" +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.normalize_ae=0')
ts.Disk.remap_config.AddLine(
    f"map http://www.ae-1.com http://127.0.0.1:{server.Variables.http_port}" +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.normalize_ae=1')
ts.Disk.remap_config.AddLine(
    f"map http://www.ae-2.com http://127.0.0.1:{server.Variables.http_port}" +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.normalize_ae=2')
# disable normalize_ae=3 on 9.1
ts.Disk.remap_config.AddLine(
    f"map http://www.ae-3.com http://127.0.0.1:{server.Variables.http_port}" +
    ' @plugin=conf_remap.so @pparam=proxy.config.http.normalize_ae=3')
ts.Disk.plugin_config.AddLine('xdebug.so --enable=x-cache')
ts.Disk.records_config.update(
    {
        'proxy.config.diags.debug.enabled': 1,
        'proxy.config.diags.debug.tags': 'http',
        'proxy.config.http.response_via_str': 3,
        # the following variables could affect the results of alternate cache matching,
        # define them with their default values explicitly
        'proxy.config.cache.limits.http.max_alts': 5,
        'proxy.config.http.cache.ignore_accept_mismatch': 2,
        'proxy.config.http.cache.ignore_accept_language_mismatch': 2,
        'proxy.config.http.cache.ignore_accept_encoding_mismatch': 2,
        'proxy.config.http.cache.ignore_accept_charset_mismatch': 2,
    })

tr = Test.AddTestRun()
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(ts)
tr.AddVerifierClientProcess("client", replay_file, http_ports=[ts.Variables.port])
tr.StillRunningAfter = ts



'''
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

Test.Summary = '''
Test ATS offering different certificates based on SNI
'''

# Define default ATS
ts = Test.MakeATSProcess("ts", enable_tls=True)
server = Test.MakeOriginServer("server", ssl=True)
dns = Test.MakeDNServer("dns")

request_header = {"headers": "GET / HTTP/1.1\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
response_header = {"headers": "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n", "timestamp": "1469733493.993", "body": ""}
server.addResponse("sessionlog.json", request_header, response_header)

# add ssl materials like key, certificates for the server
ts.addSSLfile("ssl/signed-foo.pem")
ts.addSSLfile("ssl/signed-foo.key")
ts.addSSLfile("ssl/signed-bar.pem")
ts.addSSLfile("ssl/signed2-bar.pem")
ts.addSSLfile("ssl/signed-bar.key")
ts.addSSLfile("ssl/signer.pem")
ts.addSSLfile("ssl/signer.key")
ts.addSSLfile("ssl/combo.pem")

ts.Disk.remap_config.AddLine('map / https://foo.com:{1}'.format(ts.Variables.ssl_port, server.Variables.SSL_Port))

ts.Disk.ssl_multicert_config.AddLines(
    [
        'dest_ip=127.0.0.1 ssl_cert_name=signed-foo.pem ssl_key_name=signed-foo.key',
        'ssl_cert_name=signed2-bar.pem ssl_key_name=signed-bar.key',
        'dest_ip=* ssl_cert_name=combo.pem',
    ])

# Case 1, global config policy=permissive properties=signature
#         override for foo.com policy=enforced properties=all
ts.Disk.records_config.update(
    {
        'proxy.config.ssl.server.cert.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.ssl.server.private_key.path': '{0}'.format(ts.Variables.SSLDir),
        'proxy.config.url_remap.pristine_host_hdr': 1,
        'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(dns.Variables.Port),
        'proxy.config.exec_thread.autoconfig.scale': 1.0,
        'proxy.config.dns.resolv_conf': 'NULL',
        'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
    })

dns.addRecords(records={"foo.com.": ["127.0.0.1"]})
dns.addRecords(records={"bar.com.": ["127.0.0.1"]})

# Should receive a bar.com cert
tr = Test.AddTestRun("bar.com cert")
tr.Setup.Copy("ssl/signer.pem")
tr.Setup.Copy("ssl/signer2.pem")
tr.Processes.Default.Command = "curl -v --cacert ./signer2.pem  --resolve 'bar.com:{0}:127.0.0.1' https://bar.com:{0}".format(
    ts.Variables.ssl_port)
tr.ReturnCode = 0
tr.Processes.Default.StartBefore(server)
tr.Processes.Default.StartBefore(dns)
tr.Processes.Default.StartBefore(Test.Processes.ts)
tr.StillRunningAfter = server
tr.StillRunningAfter = ts
tr.Processes.Default.Streams.All = Testers.ExcludesExpression("Could Not Connect", "Curl attempt should have succeeded")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("CN=bar.com", "Cert should contain bar.com")
tr.Processes.Default.Streams.All += Testers.ExcludesExpression("CN=foo.com", "Cert should not contain foo.com")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("404", "Should make an exchange")

# Should receive a foo.com cert
tr2 = Test.AddTestRun("foo.com cert")
tr2.Processes.Default.Command = "curl -v --cacert ./signer.pem --resolve 'foo.com:{0}:127.0.0.1' https://foo.com:{0}".format(
    ts.Variables.ssl_port)
tr2.ReturnCode = 0
tr2.StillRunningAfter = server
tr2.StillRunningAfter = ts
tr2.Processes.Default.Streams.All = Testers.ExcludesExpression("Could Not Connect", "Curl attempt should have succeeded")
tr2.Processes.Default.Streams.All += Testers.ContainsExpression("CN=foo.com", "Cert should contain foo.com")
tr2.Processes.Default.Streams.All += Testers.ExcludesExpression("CN=bar.com", "Cert should not contain bar.com")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("404", "Should make an exchange")

# Should receive random.server.com
tr2 = Test.AddTestRun("random.server.com cert")
tr2.Processes.Default.Command = "curl -v -k --resolve 'random.server.com:{0}:127.0.0.1' https://random.server.com:{0}".format(
    ts.Variables.ssl_port)
tr2.ReturnCode = 0
tr2.StillRunningAfter = server
tr2.StillRunningAfter = ts
tr2.Processes.Default.Streams.All = Testers.ExcludesExpression("Could Not Connect", "Curl attempt should have succeeded")
tr2.Processes.Default.Streams.All += Testers.ContainsExpression("CN=random.server.com", "Cert should contain random.server.com")
tr2.Processes.Default.Streams.All += Testers.ExcludesExpression("CN=foo.com", "Cert should not contain foo.com")
tr2.Processes.Default.Streams.All += Testers.ExcludesExpression("CN=bar.com", "Cert should not contain bar.com")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("404", "Should make an exchange")

# No SNI match should match specific IP address, foo.com
# SNI name and returned cert name will not match, so must use -k to avoid cert verification
tr2 = Test.AddTestRun("Bad SNI")
tr2.Processes.Default.Command = "curl -v -k --cacert ./signer.pem --resolve 'bad.sni.com:{0}:127.0.0.1' https://bad.sni.com:{0}".format(
    ts.Variables.ssl_port)
tr2.ReturnCode = 0
tr2.StillRunningAfter = server
tr2.StillRunningAfter = ts
tr2.Processes.Default.Streams.All = Testers.ExcludesExpression("Could Not Connect", "Curl attempt should have succeeded")
tr2.Processes.Default.Streams.All += Testers.ContainsExpression("CN=foo.com", "Cert should contain foo.com")
tr2.Processes.Default.Streams.All += Testers.ExcludesExpression("CN=bar.com", "Cert should not contain bar.com")
tr.Processes.Default.Streams.All += Testers.ContainsExpression("404", "Should make an exchange")



"""Verify HTTP/2 flow control behavior."""

#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from enum import Enum
from typing import List, Optional

Test.Summary = __doc__


class Http2FlowControlTest:
    """Define an object to test HTTP/2 flow control behavior."""

    _replay_file: str = 'http2_flow_control.replay.yaml'
    _replay_chunked_file: str = 'http2_flow_control_chunked.replay.yaml'
    _valid_policy_values: List[int] = list(range(0, 3))
    _flow_control_policy: Optional[int] = None
    _flow_control_policy_is_malformed: bool = False

    _default_initial_window_size: int = 65535
    _default_max_concurrent_streams: int = 100
    _default_flow_control_policy: int = 0

    _dns_counter: int = 0
    _server_counter: int = 0
    _ts_counter: int = 0
    _client_counter: int = 0

    IS_OUTBOUND = True
    IS_INBOUND = False

    IS_HTTP2_TO_ORIGIN = True
    IS_HTTP1_TO_ORIGIN = False

    class ServerType(Enum):
        """Define the type of server to use in a TestRun."""

        HTTP1_CONTENT_LENGTH = 0
        HTTP1_CHUNKED = 1
        HTTP2 = 2

    def __init__(
            self,
            description: str,
            initial_window_size: Optional[int] = None,
            max_concurrent_streams: Optional[int] = None,
            flow_control_policy: Optional[int] = None):
        """Declare the various test Processes.

        :param description: A description of the test.

        :param initial_window_size: The value with which to configure the
        proxy.config.http2.initial_window_size_(in|out) ATS parameter in the
        records.yaml file. If the paramenter is None, then no window size
        will be explicitly set and ATS will use the default value.

        :param max_concurrent_streams: The value with which to configure the
        proxy.config.http2.max_concurrent_streams_(in|out) ATS parameter in the
        records.yaml file. If the paramenter is None, then no window size
        will be explicitly set and ATS will use the default value.

        :param flow_control_policy: The value with which to configure the
        proxy.config.http2.flow_control.policy_(in|out) ATS parameter the
        records.yaml file. If the paramenter is None, then no policy
        configuration will be explicitly set and ATS will use the default
        value.
        """
        self._description = description

        self._initial_window_size = initial_window_size
        self._expected_initial_stream_window_size = (
            initial_window_size if initial_window_size is not None else self._default_initial_window_size)

        self._max_concurrent_streams = max_concurrent_streams
        self._expected_max_concurrent_streams = (
            max_concurrent_streams if max_concurrent_streams is not None else self._default_max_concurrent_streams)

        self._flow_control_policy = flow_control_policy
        self._expected_flow_control_policy = (
            flow_control_policy if flow_control_policy is not None else self._default_flow_control_policy)

        self._flow_control_policy_is_malformed = (
            self._flow_control_policy is not None and self._flow_control_policy not in self._valid_policy_values)

    def _configure_dns(self, tr: 'TestRun') -> 'Process':
        """Configure the DNS."""
        dns = tr.MakeDNServer(f'dns-{Http2FlowControlTest._dns_counter}')
        Http2FlowControlTest._dns_counter += 1
        return dns

    def _configure_server(self, tr: 'TestRun', server_type: ServerType) -> 'Process':
        """Configure the test server."""
        if server_type == self.ServerType.HTTP1_CHUNKED:
            replay_file = self._replay_chunked_file
        else:
            replay_file = self._replay_file

        server = tr.AddVerifierServerProcess(f'server-{Http2FlowControlTest._server_counter}', replay_file)
        Http2FlowControlTest._server_counter += 1
        return server

    def _configure_trafficserver(self, tr: 'TestRun', is_outbound: bool, server_type: ServerType) -> 'Process':
        """Configure a Traffic Server process."""
        ts = tr.MakeATSProcess(f'ts-{Http2FlowControlTest._ts_counter}', enable_tls=True)
        Http2FlowControlTest._ts_counter += 1

        ts.addDefaultSSLFiles()
        ts.Disk.records_config.update(
            {
                'proxy.config.ssl.server.cert.path': f'{ts.Variables.SSLDir}',
                'proxy.config.ssl.server.private_key.path': f'{ts.Variables.SSLDir}',
                'proxy.config.ssl.client.verify.server.policy': 'PERMISSIVE',
                'proxy.config.dns.nameservers': '127.0.0.1:{0}'.format(self._dns.Variables.Port),
                'proxy.config.dns.resolv_conf': 'NULL',
                'proxy.config.http.insert_age_in_response': 0,
                'proxy.config.diags.debug.enabled': 3,
                'proxy.config.diags.debug.tags': 'http',
            })

        if server_type == self.ServerType.HTTP2:
            ts.Disk.records_config.update({
                'proxy.config.ssl.client.alpn_protocols': 'h2,http/1.1',
            })

        if self._initial_window_size is not None:
            if is_outbound:
                configuration = 'proxy.config.http2.initial_window_size_out'
            else:
                configuration = 'proxy.config.http2.initial_window_size_in'
            ts.Disk.records_config.update({
                configuration: self._initial_window_size,
            })

        if self._flow_control_policy is not None:
            if is_outbound:
                configuration = 'proxy.config.http2.flow_control.policy_out'
            else:
                configuration = 'proxy.config.http2.flow_control.policy_in'
            ts.Disk.records_config.update({
                configuration: self._flow_control_policy,
            })

        if self._max_concurrent_streams is not None:
            if is_outbound:
                configuration = 'proxy.config.http2.max_concurrent_streams_out'
            else:
                configuration = 'proxy.config.http2.max_concurrent_streams_in'
            ts.Disk.records_config.update({
                configuration: self._max_concurrent_streams,
            })

        ts.Disk.ssl_multicert_config.AddLine('dest_ip=* ssl_cert_name=server.pem ssl_key_name=server.key')

        ts.Disk.remap_config.AddLine(f'map / https://127.0.0.1:{self._server.Variables.https_port}')

        if self._flow_control_policy_is_malformed:
            if is_outbound:
                configuration = 'proxy.config.http2.flow_control.policy_out'
            else:
                configuration = 'proxy.config.http2.flow_control.policy_in'
            ts.Disk.diags_log.Content = Testers.ContainsExpression(
                f"ERROR.*{configuration}", "Expected an error about an invalid flow control policy.")

        return ts

    def _configure_client(
        self,
        tr,
    ):
        """Configure a client process.

        :param tr: The TestRun to associate the client with.
        """
        tr.AddVerifierClientProcess(
            f'client-{Http2FlowControlTest._client_counter}', self._replay_file, https_ports=[self._ts.Variables.ssl_port])
        Http2FlowControlTest._client_counter += 1

    def _configure_log_expectations(self, host):
        """Configure the log expectations for the client or server."""
        hostname = "server" if host == self._server else "client"
        if self._flow_control_policy_is_malformed:
            # Since we're just testing ATS configuration errors, there's no
            # need to set up client expectations.
            return

        # ATS currently always sends a MAX_CONCURRENT_STREAMS setting.
        host.Streams.stdout += Testers.ContainsExpression(
            f'MAX_CONCURRENT_STREAMS:{self._expected_max_concurrent_streams}',
            f"{hostname} should receive a MAX_CONCURRENT_STREAMS setting.")

        if self._initial_window_size is not None:
            host.Streams.stdout += Testers.ContainsExpression(
                f'INITIAL_WINDOW_SIZE:{self._expected_initial_stream_window_size}',
                f"{hostname} should receive an INITIAL_WINDOW_SIZE setting.")

        if self._expected_flow_control_policy == 0:
            update_window_size = (self._expected_initial_stream_window_size - self._default_initial_window_size)
            if update_window_size > 0:
                host.Streams.stdout += Testers.ContainsExpression(
                    f'WINDOW_UPDATE.*id 0: {update_window_size}', f"{hostname} should receive a session WINDOW_UPDATE.")

        if self._expected_flow_control_policy in (1, 2):
            # Verify the larger window size.

            session_window_size = (self._expected_initial_stream_window_size * self._expected_max_concurrent_streams)

            # ATS will send a WINDOW_UPDATE frame to the client to increase
            # the session window size to the configured value from the default
            # value.
            update_window_size = (session_window_size - self._expected_initial_stream_window_size)

            # A WINDOW_UPDATE can only increase the window size. So make sure that
            # the new window size is greater than the default window size.
            if update_window_size > Http2FlowControlTest._default_initial_window_size:
                host.Streams.stdout += Testers.ContainsExpression(
                    f'WINDOW_UPDATE.*id 0: {update_window_size}', f"{hostname} should receive an initial session WINDOW_UPDATE.")
            else:
                # Our test traffic is large enough that eventually we should
                # send a session WINDOW_UPDATE frame for the smaller window.
                # It's not clear what it will be in advance though. A 100 byte
                # session window may not receive a 100 byte WINDOW_UPDATE frame
                # if the client is sending DATA frames in 10 byte chunks due to
                # a smaller stream window.
                host.Streams.stdout += Testers.ContainsExpression(
                    'WINDOW_UPDATE.*id 0: ', f"{hostname} should receive a session WINDOW_UPDATE.")

            if self._expected_flow_control_policy == 2:
                # Verify the streams window sizes get updated.
                stream_window_1 = session_window_size
                stream_window_2 = int(session_window_size / 2)
                stream_window_3 = int(session_window_size / 3)
                if self._server:
                    # Toward the server, there is a potential race condition
                    # between sending of first-request and the sending of the
                    # SETTINGS frame which reduces the stream window size.
                    # Allow for either scenario.
                    host.Streams.stdout += Testers.ContainsExpression(
                        (f'INITIAL_WINDOW_SIZE:{stream_window_1}.*'
                         f'INITIAL_WINDOW_SIZE:{stream_window_2}.*'),
                        f"{hostname} should stream receive window updates",
                        reflags=re.DOTALL | re.MULTILINE)
                else:
                    host.Streams.stdout += Testers.ContainsExpression(
                        (
                            f'INITIAL_WINDOW_SIZE:{stream_window_1}.*'
                            f'INITIAL_WINDOW_SIZE:{stream_window_2}.*'
                            f'INITIAL_WINDOW_SIZE:{stream_window_3}'),
                        f"{hostname} should stream receive window updates",
                        reflags=re.DOTALL | re.MULTILINE)

        if self._expected_initial_stream_window_size < 1000:
            first_id = 5 if self._server else 3

            if self._server and self._expected_flow_control_policy == 2:
                # Toward the server, there is a potential race condition
                # between sending of first-request and the sending of the
                # SETTINGS frame which reduces the stream window size. Allow
                # for either scenario.
                window_update_size = f'33|{self._expected_initial_stream_window_size}'
            else:
                window_update_size = f'{self._expected_initial_stream_window_size}'
            # For the smaller session window sizes, we expect WINDOW_UPDATE frames.
            host.Streams.stdout += Testers.ContainsExpression(
                f'WINDOW_UPDATE.*id {first_id}: {window_update_size}', f"{hostname} should receive a stream WINDOW_UPDATE.")

            host.Streams.stdout += Testers.ContainsExpression(
                f'WINDOW_UPDATE.*id {first_id + 2}: {window_update_size}', f"{hostname} should receive a stream WINDOW_UPDATE.")

            host.Streams.stdout += Testers.ContainsExpression(
                f'WINDOW_UPDATE.*id {first_id + 4}: {window_update_size}', f"{hostname} should receive a stream WINDOW_UPDATE.")

    def _configure_test_run_common(self, tr, is_outbound: bool, server_type: ServerType) -> None:
        """Perform the common Process configuration."""
        self._dns = self._configure_dns(tr)
        self._server = self._configure_server(tr, server_type)
        self._ts = self._configure_trafficserver(tr, is_outbound, server_type)
        if not self._flow_control_policy_is_malformed:
            self._configure_client(tr)
            tr.Processes.Default.StartBefore(self._dns)
            tr.Processes.Default.StartBefore(self._server)
        else:
            tr.Processes.Default.Command = "true"
        tr.Processes.Default.StartBefore(self._ts)
        tr.TimeOut = 20

    def _configure_inbound_http1_to_origin_test_run(self) -> None:
        """Configure the TestRun for inbound stream configuration."""
        tr = Test.AddTestRun(f'{self._description} - inbound, '
                             'HTTP/1 Content-Length origin')
        self._configure_test_run_common(tr, self.IS_INBOUND, self.ServerType.HTTP1_CONTENT_LENGTH)
        self._configure_log_expectations(tr.Processes.Default)

        tr = Test.AddTestRun(f'{self._description} - inbound, '
                             'HTTP/1 chunked origin')
        self._configure_test_run_common(tr, self.IS_INBOUND, self.ServerType.HTTP1_CHUNKED)
        self._configure_log_expectations(tr.Processes.Default)

    def _configure_inbound_http2_to_origin_test_run(self) -> None:
        """Configure the TestRun for inbound stream configuration."""
        tr = Test.AddTestRun(f'{self._description} - inbound, HTTP/2 origin')
        self._configure_test_run_common(tr, self.IS_INBOUND, self.ServerType.HTTP2)
        self._configure_log_expectations(tr.Processes.Default)

    def _configure_outbound_test_run(self) -> None:
        """Configure the TestRun outbound stream configuration."""
        tr = Test.AddTestRun(f'{self._description} - outbound, HTTP/2 origin')
        self._configure_test_run_common(tr, self.IS_OUTBOUND, self.ServerType.HTTP2)
        self._configure_log_expectations(self._server)

    def run(self) -> None:
        """Configure the test run for various origin side configurations."""
        self._configure_inbound_http1_to_origin_test_run()
        self._configure_inbound_http2_to_origin_test_run()
        self._configure_outbound_test_run()


#
# Default configuration.
#
test = Http2FlowControlTest("Default Configurations")
test.run()

#
# Configuring max_concurrent_streams_(in|out).
#
test = Http2FlowControlTest(description="Configure max_concurrent_streams", max_concurrent_streams=53)
test.run()

#
# Configuring initial_window_size.
#
test = Http2FlowControlTest(description="Configure a larger initial_window_size_(in|out)", initial_window_size=100123)
test.run()

#
# Configuring flow_control_policy.
#
test = Http2FlowControlTest(description="Configure an unrecognized flow_control.in.policy", flow_control_policy=23)
test.run()

test = Http2FlowControlTest(
    description="Flow control policy 0 (default): small initial_window_size",
    initial_window_size=500,  # The default is 65 KB.
    flow_control_policy=0)
test.run()
test = Http2FlowControlTest(
    description="Flow control policy 1: 100 byte session, 10 byte streams",
    max_concurrent_streams=10,
    initial_window_size=10,
    flow_control_policy=1)
test.run()
test = Http2FlowControlTest(
    description="Flow control policy 2: 100 byte session, dynamic streams",
    max_concurrent_streams=10,
    initial_window_size=10,
    flow_control_policy=2)
test.run()



#!/usr/bin/env python3
'''
Interact with a Traffic Server process.
'''
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import psutil
import signal
import sys
import traceback


class GetPidError(Exception):
    """ Raised when there was an error retrieving the specified PID."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def get_ts_process_pid(ts_identifier):
    processes = []
    for proc in psutil.process_iter(['cmdline']):
        commandline = ' '.join(proc.info['cmdline'])
        if '/traffic_server' in commandline and ts_identifier in commandline:
            return proc
    raise GetPidError("Could not find a traffic_server process")


def convert_signal_name_to_signal(signal_name):
    """
    >>> convert_signal_name_to_signal('-SIGUSR2')
    <Signals.SIGUSR2: 12>
    >>> convert_signal_name_to_signal('SIGUSR2')
    <Signals.SIGUSR2: 12>
    >>> convert_signal_name_to_signal('USR2')
    <Signals.SIGUSR2: 12>
    >>> convert_signal_name_to_signal('-USR2')
    <Signals.SIGUSR2: 12>
    >>> convert_signal_name_to_signal('KILL')
    <Signals.SIGKILL: 9>
    >>> convert_signal_name_to_signal('notasignal')
    Traceback (most recent call last):
    ...
    ValueError: Could not find a signal matching SIGnotasignal

    """
    if signal_name.startswith('-'):
        signal_name = signal_name[1:]
    if not signal_name.startswith('SIG'):
        signal_name = 'SIG' + signal_name
    for signal_value in dir(signal):
        if not signal_value.startswith('SIG'):
            continue
        if signal_name == signal_value:
            return getattr(signal, signal_value)
    raise ValueError("Could not find a signal matching {}".format(signal_name))


def parse_args():
    parser = argparse.ArgumentParser(description='Interact with a Traffic Server process')
    parser.add_argument(
        'ts_identifier', help='An identifier in the command line for the desired '
        'Traffic Server process.')
    parser.add_argument('--signal', help='Send the given signal to the process.')

    return parser.parse_args()


def main():
    args = parse_args()
    try:
        process = get_ts_process_pid(args.ts_identifier)
    except GetPidError as e:
        print(traceback.format_exception(None, e, e.__traceback__), file=sys.stderr, flush=True)
        return 1

    if args.signal:
        signal_constant = convert_signal_name_to_signal(args.signal)
        process.send_signal(signal_constant)
    else:
        print(process.pid)
    return 0


if __name__ == '__main__':
    import doctest
    doctest.testmod()
    sys.exit(main())
